﻿using System;
using Lidgren.Network;
using System.Threading;
using System.Collections.Generic;
using Jeli;

namespace Janus
{
	/// <summary>
	/// Triggered when the client successfully connects to the timeline synchronizer
	/// </summary>
	/// <param name="index">Client index number as assigned by the timeline synchronizer</param>
	public delegate void ClientConnectedHandler(ushort index);
	/// <summary>
	/// Triggered when the client disconnects from the timeline synchronizer
	/// </summary>
	/// <param name="reason">Reason for disconnection</param>
	public delegate void ClientDisconnectedHandler(string reason);

	/// <summary>
	/// Timeline client handles networking between client and the timeline server
	/// Handles connection, disconnection, timing and processing of incoming and outgoing messages
	/// </summary>
	public static class TimelineClient
	{
		static Dictionary<DeliveryMode, NetDeliveryMethod> _deliveryModes = new Dictionary<DeliveryMode, NetDeliveryMethod>()
		{
			{ DeliveryMode.ReliableOrdered, NetDeliveryMethod.ReliableOrdered },
			{ DeliveryMode.ReliableUnordered, NetDeliveryMethod.ReliableUnordered },
			{ DeliveryMode.Unreliable,NetDeliveryMethod.Unreliable }
		};

		/// <summary>
		/// Triggered when the client successfully connects to the timeline synchronizer
		/// </summary>
		public static event ClientConnectedHandler Connected;
		/// <summary>
		/// Triggered when the client disconnects from the timeline synchronizer
		/// </summary>
		public static event ClientDisconnectedHandler Disconnected;

		/// <summary>
		/// Status of the network client
		/// </summary>
		/// <returns>True if the net client is running</returns>
		public static bool IsStarted
		{
			get { return _netClient != null && _netClient.Status == NetPeerStatus.Running; }
		}

		/// <summary>
		/// Status of the network connection
		/// </summary>
		/// <returns>True if connected</returns>
		public static bool IsConnected
		{
			get
			{
				if (_netConnection != null)
					return _netConnection.Status == NetConnectionStatus.Connected;
				return false;
			}
		}

		/// <summary>
		/// Current time measured in seconds since the timeline manager started
		/// </summary>
		public static double Now
		{
			get { return TimelineManager.Default.Now; }
		}

		/// <summary>
		/// Unique index for the timeline client - assigned by the timeline synchronizer when connecting
		/// </summary>
		public static ushort Index
		{
			get { return _index; }
		}

		/// <summary>
		/// Round trip time in seconds between the client and the timeline synchronizer
		/// </summary>
		public static float RTT
		{
			get
			{
				if (_netConnection != null)
					return _netConnection.AverageRoundtripTime;
				return float.PositiveInfinity;
			}
		}

		static NetClient _netClient;
		static NetConnection _netConnection;
		static Thread _workThread;
		static bool _autoStep;
		static float _stepInterval;
		static ushort _index;
		static bool _connectSuccess;
		static bool _disconnectRequested;
		static string _disconnectReason;

		/// <summary>
		/// Start timeline client - starts separate thread that processes incoming messages
		/// </summary>
		/// <param name="autoAdd">If true, automatically connect to the timeline synchronizer using information found in TimelineClient.ini.  If false, must connect manually using Connect</param>
		/// <param name="autoStep">If true, automatically increment time.  If false, time must be manually incremented by calling Step</param>
		public static bool Start(bool autoAdd = true, bool autoStep = true)
		{
			Config config = LoadSettings();

			_autoStep = autoStep;
			_stepInterval = 1f / config["StepRate"].IntValue;

			NetPeerConfiguration netConfig = new NetPeerConfiguration("janus");
			netConfig.ConnectionTimeout = 5f;

			_netClient = new NetClient(netConfig);
			_netClient.Start();

			if (autoAdd)
			{
				Connect(config["ServerAddress"].Value, config["ServerPort"].IntValue);
			}

			_workThread = new Thread(Work);
			_workThread.Start();

			return true;
		}

		/// <summary>
		/// Get networking configuration from ini file
		/// </summary>
		static Config LoadSettings()
		{
			Config config = Config.Load("TimelineClient.ini");

			if (config == null)
			{
				config = new Config();
				config.AddChild("ServerAddress", "127.0.0.1");
				config.AddChild("ServerPort", 14242);
				config.AddChild("StepRate", 60);
				config.Save("TimelineClient.ini");
			}

			return config;
		}

		/// <summary>
		/// Connect to the timeline server
		/// </summary>
		/// <param name="serverAddress">Ip address of the server</param>
		/// <param name="serverPort">Port number on server</param>
		public static void Connect(string serverAddress, int serverPort)
		{
			_netConnection = _netClient.Connect(serverAddress, serverPort);
		}

		/// <summary>
		/// Disconnect from the timeline server
		/// </summary>
		public static void Disconnect()
		{
			_netConnection.Disconnect("");
			_netConnection = null;
			_index = 0;
		}

		/// <summary>
		/// Processes incoming and outgoing messages
		/// </summary>
		static void Update()
		{
			List<NetIncomingMessage> incomingMessages = new List<NetIncomingMessage>();
			_netClient.ReadMessages(incomingMessages);

			foreach (var message in incomingMessages)
			{
				if (message.MessageType == NetIncomingMessageType.Data)
				{
					try
					{
						TimelineMessage timelineMessage = new TimelineMessage(
							(TimelineMessageType)message.ReadByte(),
							message.ReadBytes(message.LengthBytes - 1));

						TimelineManager.Default.ProcessIncomingMessage(timelineMessage);
					}
					catch
					{
						continue;
					}
				}
				else if (message.MessageType == NetIncomingMessageType.StatusChanged)
				{
					NetConnectionStatus newStatus = (NetConnectionStatus)message.ReadByte();

					if (newStatus == NetConnectionStatus.Connected)
					{
						NetIncomingMessage remoteHail = _netConnection.RemoteHailMessage;
						_index = remoteHail.ReadUInt16();
						_connectSuccess = true;
					}
					else if (newStatus == NetConnectionStatus.Disconnected)
					{
						_connectSuccess = false;
						_disconnectRequested = true;
						_disconnectReason = message.ReadString();
					}
				}
			}

			if (_netClient.ConnectionStatus != NetConnectionStatus.Connected)
				return;

			foreach (var message in TimelineManager.Default.GetOutgoingMessages())
			{
				NetOutgoingMessage om = _netClient.CreateMessage(sizeof(byte) + message.Data.Length);
				om.Write((byte)message.MessageType);
				om.Write(message.Data);

				_netClient.SendMessage(om, _deliveryModes[message.DeliveryMode]);
			}
		}

		/// <summary>
		/// Increments the current time
		/// Triggers events for connection and disconnection
		/// </summary>
		/// <param name="deltaTime">Time step size in seconds</param>
		public static void Step (double deltaTime)
		{
			TimelineManager.Default.Step(deltaTime);
			
			if (_connectSuccess)
			{
				_connectSuccess = false;
				
				if (Connected != null)
					Connected(_index);
			}

			if (_disconnectRequested)
			{
				_connectSuccess = false;
				_disconnectRequested = false;
				_netConnection = null;
				_index = 0;
				
				if (Disconnected != null)
					Disconnected(_disconnectReason);

				_disconnectReason = null;
			}
		}

		static void Work()
		{
			DateTime lastStepTime = DateTime.Now;

			while (_netClient.Status == NetPeerStatus.Running)
			{
				Update();

				if (_autoStep)
				{
					DateTime currentTime = DateTime.Now;
					float deltaTime = (float)currentTime.Subtract(lastStepTime).TotalSeconds;

					if (deltaTime >= _stepInterval)
					{
						Step(deltaTime);
						lastStepTime = currentTime;
					}
					Thread.Sleep(1);
				}
			}
		}

		/// <summary>
		/// Stops net client
		/// </summary>
		public static void Stop()
		{
			_netConnection = null;
			_index = 0;
			_netClient.Shutdown("");

			while (_netClient.Status != NetPeerStatus.NotRunning)
			{
				Thread.Sleep(1);
			}

			_netClient = null;
		}
	}
}